

#pragma once
#ifndef CHUNK_HPP
#define CHUNK_HPP

#include <string>
#include <map>

#include "Color.hpp"
#include "Image.hpp"
#include "DataValues.hpp"

//#include "Globals.hpp"
#include "String.hpp"

#include "ChunkPrefs.hpp"

#define ANGLED_VIEW

class Chunk
{
public:
	Chunk();
	Chunk(const Chunk & chunk);
	~Chunk();

	bool Load(const STRING & file);
	bool Load(BYTE * data, UINT size);
	bool isValid(){return valid;}
	void Invalid(){valid = false;}

	void CreateImage();
	void CreateSimpleImage();
	//void CreateImageOblique();
	void CreateAngledImage();
	void CreateAngledImage2();

	Image * GetImage();

	BYTE Read(int x, int y, int z, BYTE * d);
	virtual int GetHeight(int x, int y, int z = MAPZ, int minZ = 0);
	virtual Color GetColor(int x, int y, int z, bool side = false);
	Color RayCast(int fromX, int fromY, int fromZ, int toX, int toY, int toZ);

	inline COORDS GetX(){return chunkX;}
	inline COORDS GetY(){return chunkY;}
	inline INT64 GetLastUpdate(){return lastUpdate;}

	// Count what is in the map
	struct Counter
	{
		Counter();
		Counter & operator=(const Counter &c);
		Counter & operator+=(const Counter &c);
		const Counter & operator+(const Counter &c) const;
		INT64 
		// Blocks
			block[BLOCKAMOUNT_MAX];
		// Mobs
		std::map<std::string, INT64> mobs;
			/*chicken,
			sheep,
			zombie,
			pig,
			skeleton,
			creeper,
			cow,
			slime,
			pigzombie,
			ghast;*/
		size_t filesize;
	} amount;

	ChunkPrefs pref;
	STRING file;

protected:
	virtual void Allocate();
	virtual void Deallocate();

	virtual bool ParseNBT(void *);

	BYTE * blocks;
	BYTE * blockData;
	BYTE * skyLight;
	BYTE * blockLight;
	BYTE * heightMap;

	COORDS chunkX, chunkY;
	INT64 lastUpdate;

	bool valid;
	
	Image * image;

	// Function passes
	virtual void PassExclude(int & x, int & y, int & z, Color & color);
	void PassNormal(int & x, int & y, int & z, Color & color);
	void PassGray(int & x, int & y, int & z, Color & color);
	void PassColor(int & x, int & y, int & z, Color & color);
	void PassCave(int & x, int & y, int & z, Color & color);
	void PassBlend(int & x, int & y, int & z, Color & color);
	void PassHeight(int & x, int & y, int & z, Color & color);
	virtual void PassNight(int & x, int & y, int & z, Color & color);
};

#endif // CHUNK_HPP