

#pragma once
#ifndef CHUNKPREFS_HPP

#include <string.h>

#include "DataValues.hpp"

#include "Transform.hpp"

#include "Number.hpp"

// Map size
#define MAPX 16
#define MAPY 16
#define MAPZ 128
#define ANVILZ 256
const int MAXBLOCKS = MAPX * MAPY * MAPZ;
const int SECTIONBLOCKS = MAPX * MAPY * 16;
const int MAXSECTIONS = ANVILZ / 16;
#define MAPCOORD 10000000

// Flags
#define CHUNKP_NIGHT		(1 << 0)
#define CHUNKP_NORENDER		(1 << 1)
#define CHUNKP_CAVE			(1 << 2)
#define CHUNKP_COLOR		(1 << 3)
#define CHUNKP_GRAY			(1 << 4)
#define CHUNKP_SOLID		(1 << 5)
#define CHUNKP_NOHEIGHTMAP	(1 << 6)
#define CHUNKP_SINGLESLICE	(1 << 7)
#define CHUNKP_INTERPOLATION (1 << 8)

// Map version
// Note: The old ones are only there if there will be a backcompatibility.
// They might be added later if requested.
enum MAP_VERSION
{
	UNKNOWN,
	CLASSIC,
	INVDEV,
	ALPHA,
	BETA,
	DISCOVER,
	ANVIL
};

// Preferences for a chunk
struct ChunkPrefs
{
	ChunkPrefs()
		: version(UNKNOWN), flags(0), z(MAPZ), h(0), logging(false), rotation(0)
	{
		memset(block, true, BLOCKAMOUNT_MAX);
	}
	ChunkPrefs & operator=(const ChunkPrefs & pref)
	{
		memcpy(this, &pref, sizeof(ChunkPrefs));
		return *this;
	}
	MAP_VERSION version;
	union
	{
		// Bitfield is used to make it easier to change each flag
		// Please note that this wont work on certain compilers
		struct
		{
			bool night : 1;
			bool norender : 1;
			bool cave : 1;
			bool color : 1;
			bool gray : 1;
			bool solid : 1;
			bool noheightmap : 1;
			bool singleslice : 1;
			bool interpolation : 1;
		} flags_s;
		UINT flags;
	};
	int z;
	BYTE h;
	bool block[BLOCKAMOUNT_MAX];
	bool logging;
	int rotation;
	Rotation rot;
};

#endif // CHUNKPREFS_HPP