

#pragma once
#ifndef IMAGE_HPP

#include "Color.hpp"
//#include "Globals.hpp"
#include "String.hpp"
#include "Number.hpp"

#define PNG_DEBUG 3
#ifdef WIN64
#include "win64/png.h"
#elif defined(WIN32)
#include "win32/png.h"
#elif defined(UNIX)
#include <png.h>
#endif

struct ImageSave
{
public:
	ImageSave() : save(0) {};
	bool Prepare(STRING file, UINT width, UINT height);
	bool Row(Color * data);
	bool Close();
private:
	FILE * save;
	png_structp png;
	png_infop info;
};

struct Image
{
public:
	Image();
	Image(const Image & image);
	Image(UINT width, UINT height);
	~Image();

	inline UINT Width() {return width;}
	inline UINT Height() {return height;}
	void Resize(int x1, int y1, int x2, int y2);

	bool Load(STRING file);
	bool Save(STRING file);

	// Save to image for each row
	bool PrepareSave(STRING file);
	bool SaveRow(Color * data);
	bool SaveRow(UINT y);
	bool CloseSave();

	void * GetRawData();
	void SetPixel(UINT x, UINT y, Color pixel);
	void SetPixels(double x, double y, Color pixel);
	Color GetPixel(UINT x, UINT y);
	void Draw(UINT x, UINT y, Image * img);

	Color Average();
private:
	Color * data;
	UINT width, height;

	// When saving in rows
	ImageSave * rows;
};

#endif // IMAGE_HPP