

#include <stdlib.h>

#include "Region.hpp"
#include "nbt.hpp"

#include "Level.hpp"
#include "ListFiles.hpp"

#include "ThreadPool.hpp"

//#include "Utility.hpp"
#include "System.hpp"
//#include "Computer.hpp"

#include <algorithm>

using namespace std;

// Thread struct
/*struct ThreadString
{
	STRING file;
	ChunkPrefs prefs;
	Chunk * chunk;
};*/

// Working thread for the pool
RESULT ThreadPoolWork(WORK order)
{
	Chunk * chunk = (Chunk *)order;
	//ThreadString * next = (ThreadString *)order;
	//ChunkPrefs prefs = next->prefs;
	//Chunk * chunk = next->chunk;
	//chunk->pref = prefs;
	STRING file = chunk->file;
	// This SHOULD increase performance
	//delete next;

	// Alpha
	if (chunk->pref.version == ALPHA)
	{
		chunk->Load(file);
	}
	// Beta
	else if (chunk->pref.version == BETA || chunk->pref.version == ANVIL)
	{
		chunk->Load((BYTE *)file.c_str(), (UINT)file.size());
	}
	if (chunk->isValid())
	{
		chunk->CreateImage();
		return (RESULT)1;
	}
	return (RESULT *)0;
}

// Constructor
Level::Level()
{
	Reset();
}

// Destructor
Level::~Level()
{
	ClearChunks();
}

// Abort
void Level::Abort()
{
	cancel = true;
	if (state == DIRECTORY)
		total = -1;
}

// Reset everything
void Level::Reset()
{
	total = 0;
	work = 0;
	result = 0;
	state = INITIALIZE;
	done = 0;
	saved = false;
	cancel = false;

	blocks = 0;
}

// Clear all chunks
void Level::ClearChunks()
{
	/*while (!chunks.empty())
	{
		Chunk * t = chunks.back();
		delete t;
		chunks.pop_back();
	}*/
	chunks.clear();
}

// Load level information
int Level::LoadLevel(const STRING & path)
{
	STRING file;
	file.assign(path.c_str());
	STRING name = Port::FileName(file);
	file.append(STR("/level.dat"));
	NBT_Reader read = NBT_Reader();
	read.Load(Port::ReplacePath(file));

	bool valid = false;

	NBTTag tag = read.ReadData(); // Read first compound
	while ((tag = read.ReadData()) != TAG_Unknown)
	{
		const NBTData * data = read.GetData();
		if (!data)
			continue;
		if (!valid)
		{
			if (data->tag != TAG_Compound)
				continue;
			if (!data->name)
				continue;
			if (!StringCheck(data->name, "Data"))
				continue;
			valid = true;
		}
		switch (data->tag)
		{
			case TAG_Int:
				if (StringCheck(data->name, "version"))
				{
					info.version = data->i;
				}
				break;
			case TAG_Long:
				if (StringCheck(data->name, "RandomSeed"))
				{
					info.seed = data->l;
				}
				break;
			case TAG_String:
				if (StringCheck(data->name, "LevelName"))
				{
					info.name.assign((char *)data->bas, (char *)data->bas+data->size);
				}
				break;
		}
	}
	// Alpha
	if (info.name.empty())
		info.name.assign(name.begin(), name.end());
	return valid;
}

// Load a world folder
int Level::LoadWorld(const STRING & p)
{
	if (cancel)
		return LVL_ERR_CANCEL;
	Reset();
	state = DIRECTORY;
	STRING path = p;

	STRING search;
	{
		char t[] = {"*"};
		search.assign(t, t+strlen(t));
	}

	// Find files
	Files fs(path, true);
	fs.exclude.push_back(STR("DIM-1"));
	STRING file;
	while (fs.NextFile(file) && !cancel)
	{
		if(prefs.version == BETA)
		{
			if(file.length() > 3 && file.substr(file.length() - 3, 3).compare("mcr") == 0)
			{
				files.push(file);
				total++;
			}
		}
		else if(prefs.version == ANVIL)
		{
			if(file.length() > 3 && file.substr(file.length() - 3, 3).compare("mca") == 0)
			{
				files.push(file);
				total++;
			}
		}
	}

	if (cancel)
		return LVL_ERR_CANCEL;
	total = files.size();
	work = 0;
	result = 0;
	done = 0;
	prefs.logging = !prefs.log.empty();
	// Get correct amounts of chunks
	if (prefs.version == BETA || prefs.version == ANVIL)
	{
		total = 0;
		stack<STRING> tempFiles;
		while (!files.empty())
		{
			STRING str = files.top();
			files.pop();
			RegionReader region;
			if (region.Load(str))
			{
				region.ParseHeader();
				total += region.GetAmountChunks();
				tempFiles.push(str);
			}
		}
		files = tempFiles;
	}
	state = PROCESSING;

	// Allocate
	chunks.reserve(total);

	// Limit threads
	if (prefs.threads > files.size())
		prefs.threads = files.size();
	// More threads than the computer can handle is no good for you
	size_t i = 0;
	if (prefs.threads > (i = Port::GetLogicalCores()))
		prefs.threads = i & 0xFF;
	// Multithreaded program
	if (prefs.threads >= 2)
	{
		LoadMultiple();
	}
	else
	{
		LoadSingle();
	}

	// Mark cleanup progress
	state = CLEANING;

	// Fastest way to clear a stack
	while (!files.empty())
		files.pop();

	if (cancel)
		return LVL_ERR_CANCEL;

	return LVL_OK;
}

// Load singlethreaded
void Level::LoadSingle()
{
	// Go through all files
	Chunk * temp = 0;
	while (files.empty() == false && cancel == false)
	{
		if (prefs.version == ALPHA)
		{
			temp = new Chunk();
			// Load chunk
			temp->pref = prefs;
			temp->Load(files.top());
			files.pop();
			// Everything is fine
			if (temp->isValid())
			{
				temp->CreateImage();
				chunks.push_back(temp);
			}
			else
			{
				// Display correct total chunks
				--total;
				// Evade memory leak
				delete temp;
			}
			// Do some aftermath
			result = chunks.size();
			work = total - result;
			done = ((float)result/(float)total);
		}
		else if (prefs.version == BETA || prefs.version == ANVIL)
		{
			RegionReader region = RegionReader();
			region.Load(files.top());
			files.pop();
			region.ParseData();
			// Go through region
			for (size_t i = 0; i < region.GetAmountChunks() && cancel == false; ++i)
			{
				if(prefs.version == BETA)
					temp = new Chunk();
				else if(prefs.version == ANVIL)
					temp = new AnvilChunk();
				// Load chunk
				temp->pref = prefs;
				if(temp->pref.version == ANVIL)
					temp->pref.z = ANVILZ;
				UINT size = 0;
				BYTE * data = region.GetChunk(i, size);
				temp->Load(data, size);
				// Everything is fine
				if (temp->isValid())
				{
					const double PI = 3.14159265;
					temp->CreateImage();
					chunks.push_back(temp);
				}
				else
				{
					// Display correct total chunks
					--total;
					// Evade memory leak and decrease time to clean up
					delete temp;
				}
				// Do some aftermath
				result = chunks.size();
				work = total - result;
				done = ((float)result/(float)total);
			}
		}
	}
	
	// Clean up if canceled
	if (cancel)
	{
		state = CLEANING;
		// Remove chunks
		while (!chunks.empty())
		{
			Chunk * chunk = chunks.back();
			delete chunk;
			chunks.pop_back();
		}
	}
}

// Load with multithreaded
void Level::LoadMultiple()
{
	ThreadPool * pool = new ThreadPool();
	pool->Initialize(prefs.threads, ThreadPoolWork);

	// Preparing region object
	RegionReader region = RegionReader();
	int current = 0;
	bool useRegion = (prefs.version == BETA || prefs.version == ANVIL);

	int n = 0;

	// Start swimming
	do 
	{
		// Get latest result
		RESULT chnk = 0;
		if (pool->Pop(&chnk))
		{
			if (chnk != 0)
			{
				++result;
			}
			else
			{
				--total;
			}
		}
		// Push work
		if (useRegion)
		{
			if (files.empty() && region.GetAmountChunks() == current)
				useRegion = false;
			else if (!files.empty() && (size_t)current >= region.GetAmountChunks())
			{
				region.Load(files.top());
				files.pop();
				region.ParseData();
				current = 0;
			}
			if ((size_t)current < region.GetAmountChunks())
			{
				UINT size = 0;
				BYTE * data = region.GetChunk(current++, size);
				Chunk * chunk = new Chunk();
				chunk->file.assign(data, data+size);
				chunk->pref = prefs;
				chunks.push_back(chunk);
				pool->Push((WORK)chunk);
				++n;
			}
		}
		else if (!files.empty())
		{
			Chunk * chunk = new Chunk();
			chunk->file = files.top();
			chunk->pref = prefs;
			chunks.push_back(chunk);
			pool->Push((WORK)chunk);

			// Pop things
			files.pop();
			++n;
		}
		work = (total - files.size()) - result;
		done = ((float)result/(float)total);
	}
	while (pool->Work() && !cancel);

	if (cancel)
	{
		pool->Shutdown();
		state = CLEANING;
		// Remove chunks
		while (!chunks.empty())
		{
			delete chunks.back();
			chunks.pop_back();
		}
	}

	// Clean up after you
	pool->Shutdown();

#ifdef WIN32
	if (pool->GetClosedThreads() < (int)prefs.threads)
		MessageBox(0, TEXT("Some threads were not released."), TEXT("PixelMap"), MB_OK);
#endif

	delete pool;
}

#include <math.h>

#include "FileCache.hpp"

Rotation g_transform;
INT64 minX = MAPCOORD;
INT64 minY = MAPCOORD;
INT64 maxX = -MAPCOORD;
INT64 maxY = -MAPCOORD;

// Only way to do this due of how vector works
int amountDuplicates = 0;

// Make for std::sort
bool compare_chunks(Chunk * chnk1, Chunk * chnk2)
{
	if (chnk1 == 0)
		return false;
	if (chnk2 == 0)
		return true;
	Vector3D a(double((chnk1->GetX() - minX) * MAPX - ((double)g_transform.width / 2)), double((chnk1->GetY() - minY) * MAPY - ((double)g_transform.height / 2)), 0);
	Vector3D b(double((chnk2->GetX() - minX) * MAPX - ((double)g_transform.width / 2)), double((chnk2->GetY() - minY) * MAPY - ((double)g_transform.height / 2)), 0);
	a = g_transform.transform * a;
	b = g_transform.transform * b;

	if (a.y < b.y)
		return true;
	if (a.y > b.y)
		return false;
	if (a.x > b.x)
		return true;
	if (a.x < b.x)
		return false;
	++amountDuplicates;
	return false;
}

// Create image saved in file
int Level::CreateImage(const STRING & file)
{
	if (cancel)
		return LVL_ERR_CANCEL;
	state = CREATEIMAGE;
	if (chunks.empty())
		return LVL_ERR_NOCHUNKS;

	blocks = chunks.size() << 8;
	bool doRender = !(prefs.flags & CHUNKP_NORENDER);
	minX = MAPCOORD;
	minY = MAPCOORD;
	maxX = -MAPCOORD;
	maxY = -MAPCOORD;

	CalculateSize();

	if (cancel)
		return LVL_ERR_CANCEL;

	if (doRender)
	{
		// No valid chunks?
		if (minX == MAPCOORD && minY == MAPCOORD &&
			maxX == -MAPCOORD && maxY == -MAPCOORD)
			return LVL_ERR_INVALIDCHUNKS;

		prefs.rotation = abs((prefs.rotation / 90) * 90) % 360;

		UINT
			maxWidth = (UINT)(maxX - minX),
			maxHeight = (UINT)(maxY - minY);

		UINT
			width = (maxWidth + 1) * MAPX, 
			height = (maxHeight + 1) * MAPY;

		// Foolproof
		if (width == 0 || height == 0)
			return LVL_ERR_INVALIDSIZE;

		// Rotate
		
		//MATRIX trans;
		Rotation rot = GetRotation(width, height);
		
/*#ifdef ANGLED_VIEW
		MATRIX trans;
		Rotation rot;
		{
			rot.width = width;
			rot.height = height;
			rot.depth = MAPZ;
			rot.Transform(prefs.yaw, prefs.pitch);
			width = rot.width;
			height = rot.height;
			trans = rot.transform;
			g_transform = rot;
		}
#else
		{
			float rCos = cos(rad);
			float rSin = sin(rad);
			rCos = (rCos < 0) ? -rCos : rCos;
			rSin = (rSin < 0) ? -rSin : rSin;
			UINT newWidth = UINT(width * rCos + height * rSin);
			UINT newHeight = UINT(height * rCos + width * rSin);
			width = newWidth;
			height = newHeight;
		}
#endif*/

		// Foolproof
		if (width == 0 || height == 0)
			return LVL_ERR_INVALIDSIZE;

		bool useCache = !prefs.cache.empty();
		ImageCache cache(width, height);
		Image * image = 0;
		int dMinX = 0, dMinY = 0, dMaxX = 0, dMaxY = 0;

		int error = PickCacheType(&image, &cache, width, height);
		if (error != LVL_OK)
			return error;

		// Sort list
//#ifdef ANGLED_VIEW
		bool isStraight = prefs.rot.IsStraight();
		if (!isStraight)
			//chunks.sort(compare_chunks);
			std::sort(chunks.begin(), chunks.end(), compare_chunks);
//#endif
		// Found duplicates
		//if (amountDuplicates > 0)
		//	return LVL_ERR_DUPLICATES;

		// Create image
		int n = 0;
		for (vector<Chunk *>::iterator i = chunks.begin(); i != chunks.end(); ++i, ++n)
		{
			Chunk * chunk = (*i);
			if (!chunk->isValid())
				continue;

			Image * img = chunk->GetImage();
			// One more try
			if (!img)
			{
				chunk->pref = prefs;
				chunk->CreateImage();
				img = chunk->GetImage();
				if (!img)
					continue;
			}

			//if (prefs.rot.width == MAPX && prefs.rot.height == MAPY)
			// Default rotation
			if (isStraight)
			{
				UINT x = 0;
				UINT y = 0;
				switch (prefs.rotation)
				{
					case 90:
						x = (width - (UINT(chunk->GetY() - minY) * (MAPY))) - MAPY;
						y = UINT(chunk->GetX() - minX) * MAPX;
						break;
					case 180:
						x = (width - (UINT(chunk->GetX() - minX) * (MAPX))) - MAPX;
						y = (height - (UINT(chunk->GetY() - minY) * (MAPY))) - MAPY;
						break;
					case 270:
						x = UINT(chunk->GetY() - minY) * MAPY;
						y = (height - (UINT(chunk->GetX() - minX) * (MAPX))) - MAPX;
						break;
					case 0:
					default:
						x = UINT(chunk->GetX() - minX) * MAPX;
						y = UINT(chunk->GetY() - minY) * MAPY;
				}

				Color chnk[MAPX][MAPY] = {0};

				// Write pixel for pixel
				for (UINT X = 0; X < MAPX; ++X)
				{
					for (UINT Y = 0; Y < MAPY; ++Y)
					{
						UINT XX = 0;
						UINT YY = 0;
						switch (prefs.rotation)
						{
							case 90:
								XX = Y;
								YY = (MAPX - 1) - X;
								break;
							case 180:
								XX = (MAPX - 1) - X;
								YY = (MAPY - 1) - Y;
								break;
							case 270:
								XX = (MAPY - 1) - Y;
								YY = X;
								break;
							case 0:
							default:
								XX = X;
								YY = Y;
						}
						if (!useCache)
						{
							image->SetPixel((height - (y + Y)) - 1, (x + X), img->GetPixel(YY, XX));
						}
						else
						{
							Color * current = (*chnk) + ((MAPY - (Y+1)) + (MAPY * X));
							Color c = img->GetPixel(YY, XX);
							memcpy(current, &c, sizeof(Color));
						}
					}
				}
				if (useCache)
				{
					for (UINT X = 0; X < MAPX; ++X)
					{
						cache.Write(chnk[X], (height - y) - MAPY, (x + X), MAPY);
					}
				}
			}
			// Angled rotation
			else
			{
				Vector3D pos = rot.Place(double((chunk->GetX() - minX) * MAPX), double((chunk->GetY() - minY) * MAPY), 0);
				//Vector3D pos(double((chunk->GetX() - minX) * MAPX - ((double)width / 2)), double((chunk->GetY() - minY) * MAPY - ((double)height / 2)), 0);
				//pos = trans * pos;
				int x = abs((int)pos.x);
				int y = abs((int)pos.y);
				image->Draw(x, y, img);
				if (x < dMinX)
					dMinX = x;
				if (y < dMinY)
					dMinY = y;
				if (x > dMaxX)
					dMaxX = x;
				if (y > dMaxY)
					dMaxY = y;
			}
/*#ifdef ANGLED_VIEW
			Vector3D pos = rot.Place(double((chunk->GetX() - minX) * MAPX), double((chunk->GetY() - minY) * MAPY), 0);
			//Vector3D pos(double((chunk->GetX() - minX) * MAPX - ((double)width / 2)), double((chunk->GetY() - minY) * MAPY - ((double)height / 2)), 0);
			//pos = trans * pos;
			int x = abs((int)pos.x);
			int y = abs((int)pos.y);
			image->Draw(x, y, img);
			if (x < dMinX)
				dMinX = x;
			if (y < dMinY)
				dMinY = y;
			if (x > dMaxX)
				dMaxX = x;
			if (y > dMaxY)
				dMaxY = y;
#else // ANGLED_VIEW
			UINT x = 0;
			UINT y = 0;
			switch (prefs.rotation)
			{
				case 90:
					x = (width - (UINT(chunk->GetY() - minY) * (MAPY))) - MAPY;
					y = UINT(chunk->GetX() - minX) * MAPX;
					break;
				case 180:
					x = (width - (UINT(chunk->GetX() - minX) * (MAPX))) - MAPX;
					y = (height - (UINT(chunk->GetY() - minY) * (MAPY))) - MAPY;
					break;
				case 270:
					x = UINT(chunk->GetY() - minY) * MAPY;
					y = (height - (UINT(chunk->GetX() - minX) * (MAPX))) - MAPX;
					break;
				case 0:
				default:
					x = UINT(chunk->GetX() - minX) * MAPX;
					y = UINT(chunk->GetY() - minY) * MAPY;
			}

			Color chnk[MAPX][MAPY] = {0};

			// Write pixel for pixel
			for (UINT X = 0; X < MAPX; ++X)
			{
				for (UINT Y = 0; Y < MAPY; ++Y)
				{
					UINT XX = 0;
					UINT YY = 0;
					switch (prefs.rotation)
					{
						case 90:
							XX = Y;
							YY = (MAPX - 1) - X;
							break;
						case 180:
							XX = (MAPX - 1) - X;
							YY = (MAPY - 1) - Y;
							break;
						case 270:
							XX = (MAPY - 1) - Y;
							YY = X;
							break;
						case 0:
						default:
							XX = X;
							YY = Y;
					}
					if (!useCache)
					{
						image->SetPixel((height - (y + Y)) - 1, (x + X), img->GetPixel(YY, XX));
					}
					else
					{
						Color * current = (*chnk) + ((MAPY - (Y+1)) + (MAPY * X));
						Color c = img->GetPixel(YY, XX);
						memcpy(current, &c, sizeof(Color));
					}
				}
			}
			if (useCache)
			{
				for (UINT X = 0; X < MAPX; ++X)
				{
					cache.Write(chnk[X], (height - y) - MAPY, (x + X), MAPY);
				}
			}
#endif // ANGLED_VIEW*/

			done = ((float)n/(float)total);

			delete chunk;
		}
		done = 1;

		// No render mode
		if (prefs.flags & CHUNKP_NORENDER)
		{
			saved = true;
			if (!useCache)
			{
				delete image;
			}
			else
			{
				cache.Close();
			}
			state = FINALIZING;
			return LVL_OK;
		}
		// Save image
		state = SAVING;

		// Normal save of image
		if (!useCache)
		{
			if (image->PrepareSave(file))
			{
				for (UINT i = 0; i < width; ++i)
				{
					if (!image->SaveRow(i))
						break;
					done = (float)i/(float)width;
				}
				saved = image->CloseSave();
			}
			done = 1;
			delete image;
		}
		// Save from cache
		else
		{
			ImageSave sav;
			if (sav.Prepare(file, height, width))
			{
				Color * line = new Color[height];
				for (UINT i = 0; i < width; ++i)
				{
					cache.Read(&line, 0, i, height);
					if (!sav.Row(line))
						break;
					done = (float)i/(float)width;
				}
				delete[] line;
				saved = sav.Close();
			}
			cache.Close();
			done = 1;
		}
	}
	else
	{
		saved = true;
	}
	state = FINALIZING;
	return saved ? LVL_OK : LVL_ERR_SAVE;
}

void Level::CalculateSize()
{
	bool doRender = !(prefs.flags & CHUNKP_NORENDER);

	// Calculate size
	for (vector<Chunk *>::iterator i = chunks.begin(); i != chunks.end(); ++i)
	{
		Chunk * chunk = (*i);
		if (!chunk->isValid())
			continue;
		if (doRender)
		{
			COORDS x = chunk->GetX();
			COORDS y = chunk->GetY();
			if (x < minX)
				minX = x;
			if (y < minY)
				minY = y;
			if (x > maxX)
				maxX = x;
			if (y > maxY)
				maxY = y;
		}

		// And add to amount while we are processing
		amount += chunk->amount;
	}
}

Rotation Level::GetRotation(UINT & width, UINT & height)
{
	float rad = (2 * 3.14159265f * prefs.rotation) / 360;
	Rotation rot;
	if (!prefs.rot.IsStraight())
	{
		rot.width = width;
		rot.height = height;
		rot.depth = MAPZ;
		rot.Transform(prefs.yaw, prefs.pitch);
		width = rot.width;
		height = rot.height;
		//trans = rot.transform;
		g_transform = rot;
	}
	else
	{
		float rCos = cos(rad);
		float rSin = sin(rad);
		rCos = (rCos < 0) ? -rCos : rCos;
		rSin = (rSin < 0) ? -rSin : rSin;
		UINT newWidth = UINT(width * rCos + height * rSin);
		UINT newHeight = UINT(height * rCos + width * rSin);
		width = newWidth;
		height = newHeight;
	}
	return rot;
}

int Level::PickCacheType(Image ** image, ImageCache * cache, UINT width, UINT height)
{
// Linux supportive
#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

	if (prefs.cache.empty())
	{
		// No more than 1 GiB
		if (width * height * sizeof(Color) > min(1073741824, Port::GetAvailableMemory()))
			return LVL_ERR_TOOLARGE;
		try
		{
			*image = new Image(width, height);
			if (*image == 0)
				throw LVL_ERR_TOOLARGE;
		}
		catch (Level_Error error)
		{
			return error;
		}
	}
	else
	{
		STRING cacheFile = prefs.cache;
		//cacheFile.append(STR("\\cache.bin"));
		//cacheFile = Port::ReplacePath(cacheFile);

		if (!cache->Open(cacheFile))
			return LVL_ERR_CACHE;
	}
	return LVL_OK;
}