

#include "ListFiles.hpp"

//#include "Utility.hpp"
#include "String.hpp"

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#elif defined(UNIX)
#include <sys/types.h>
#include <dirent.h>
#endif // WIN32 / UNIX

// A lot easier
#ifdef UNICODE
#define STRCMP wcscmp
#else
#define STRCMP strcmp
#endif

using namespace std;

FileSearch::FileSearch()
{
	exclude.push_back(STR("."));
	exclude.push_back(STR(".."));
	amount = 0;
	deep = true;
}

// List files
bool ListFiles(FileSearch * prefs)
{
	if (!prefs)
		return false;
	// Clear strings
	prefs->path.assign(prefs->path.c_str());

	stack<STRING> directories;
	STRING file;
	STRING path = prefs->path;
	directories.push(path);

	vector<STRING> * exclude = &prefs->exclude;

	size_t * amount = prefs->amount;

#ifdef WIN32
	HANDLE find = INVALID_HANDLE_VALUE;

	WIN32_FIND_DATA foundData = {0};

	while (!directories.empty())
	{
		path = directories.top();
		directories.pop();
		file = path + STR("\\*");
		file.assign(file.c_str());

		find = FindFirstFile(file.c_str(), &foundData);
		if (find == INVALID_HANDLE_VALUE)
			return false;

		do
		{
			if (amount && (*amount) == -1)
				return false;
			bool found = false;
			for (vector<STRING>::iterator i = exclude->begin(); i != exclude->end(); ++i)
			{
				if (STRCMP(foundData.cFileName, i->c_str()) == 0)
				{
					found = true;
					break;
				}
			}
			if (found)
				continue;
			
			if (prefs->deep && foundData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				directories.push(path + STR("\\") + foundData.cFileName);
			else
				prefs->files.push(path + STR("\\") + foundData.cFileName);
			if (amount)
				(*amount) = prefs->files.size();
		}
		while (FindNextFile(find, &foundData) != 0);

		FindClose(find);
		if (GetLastError() != ERROR_NO_MORE_FILES)
			return false;

		find = INVALID_HANDLE_VALUE;
	}
#elif defined(UNIX)
	DIR * dp = 0;
	dirent * dirp = 0;

	while (!directories.empty())
	{
		path = directories.top();
		directories.pop();
		file = path;

		dp = opendir(path.c_str());
		if (!dp)
			continue;

		while (dirp = readdir(dp))
		{
			if (amount && (*amount) == -1)
			{
				closedir(dp);
				return false;
			}
			bool found = false;
			for (vector<STRING>::iterator i = exclude->begin(); i != exclude->end(); ++i)
			{
				if (STRCMP(dirp->d_name, i->c_str()) == 0)
				{
					found = true;
					break;
				}
			}
			if (found)
				continue;
			if (prefs->deep && dirp->d_type == DT_DIR)
				directories.push(path + STR("/") + dirp->d_name);
			else if (dirp->d_type == DT_REG)
				prefs->files.push(path + STR("/") + dirp->d_name);
			// Update amount
			if (amount)
				(*amount) = prefs->files.size();
		}
		closedir(dp);
	}
#endif // WIN32 / UNIX

	return true;
}

// Holder for data
struct FilesData
{
	FilesData()
	{
// Windows
#ifdef WIN32
		find = INVALID_HANDLE_VALUE;
		memset(&foundData, 0, sizeof(foundData));
// Unix
#elif defined(UNIX)
		dp = 0;
		dirp = 0;
#endif // WIN32 / UNIX
	}

	~FilesData()
	{
// Windows
#ifdef WIN32
		if (find != INVALID_HANDLE_VALUE)
			FindClose(find);
// Unix
#elif defined(UNIX)
		if (dp)
			closedir(dp);
#endif // WIN32 / UNIX
	}

	const STRING GetFile()
	{
#ifdef WIN32
		return foundData.cFileName;
#elif defined(UNIX)
		return dirp->d_name;
#endif
		return "";
	}

// Windows
#ifdef WIN32
	HANDLE find;
	WIN32_FIND_DATA foundData;
// Unix
#elif defined(UNIX)
	DIR * dp;
	dirent * dirp;
#endif // WIN32 / UNIX
};

typedef FilesData * FD;

// Constructor
Files::Files(STRING p, bool d)
{
	exclude.push_back(STR("."));
	exclude.push_back(STR(".."));
	amount = 0;
	sub = d;
	startPath = p;
	currentPath = startPath;
	directories.push(currentPath);

	data = new FilesData();
}

// Destructor
Files::~Files()
{
	delete (FD)data;
	data = 0;
}

// Get next file
bool Files::NextFile(STRING & returnFile)
{
	STRING file;

	if (!FindNext())
		return false;

	do
	{
		// Excluded
		if (IsExcluded())
		{
			if (!FindNext())
				return false;
			continue;
		}

		// Found a file
		if (FoundFile(returnFile))
			return true;

		// Continue the search
		if (!FindNext())
			return false;
	}
	// Go around until something happens
	while (true);

	// Just because
	return true;
}

// Find next directory
bool Files::OpenNextDir()
{
	FD d = (FD)data;
	if (directories.empty())
		return false;

	// Prepare next path
	currentPath = directories.top();
	directories.pop();

	currentFile = currentPath + STR("\\*");
	currentFile.assign(currentFile.c_str());

#ifdef WIN32
	d->find = FindFirstFile(currentFile.c_str(), &d->foundData);
#elif defined(UNIX)
	d->dp = opendir(currentFile.c_str());
#endif
	// Something happened, so we break here
	if (!IsDirOpen())
		return false;
	return true;
}

// Find next file
bool Files::FindFile()
{
	FD d = (FD)data;
#ifdef WIN32
	if (!FindNextFile(d->find, &d->foundData))
	{
		FindClose(d->find);
		//if (GetLastError() == ERROR_NO_MORE_FILES)
		d->find = INVALID_HANDLE_VALUE;
		return false;
	}
#elif defined(UNIX)
	if (!(d->dirp = readdir(d->dp)))
	{
		closedir(d->dp);
		d->dp = 0;
		return false;
	}
#endif
	return true;
}

// Checks all directories after files
bool Files::FindNext()
{
	// Next directory
	if (!IsDirOpen())
	{
		while (!OpenNextDir())
		{
			if (directories.empty())
				return false;
		}
	}
	// Find file
	else
	{
		while (!FindFile())
		{
			if (directories.empty())
				return false;
			// Recursive
			else
				return FindNext();
		}
	}
	return true;
}

// Check if a file was found
bool Files::FoundFile(STRING & returnFile)
{
	FD d = (FD)data;
#ifdef WIN32
	// Directory
	if (sub && d->foundData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	{
		directories.push(currentPath + STR("\\") + d->foundData.cFileName);
		return false;
	}
	// File found
	else
	{
		returnFile = currentPath + STR("\\") + d->foundData.cFileName;
		++amount;
		return true;
	}
#elif defined(UNIX)
	// Directory
	if (sub && d->dirp->d_type == DT_DIR)
	{
		directories.push(currentPath + STR("/") + d->dirp->d_name);
		return false;
	}
	// File found
	else if (d->dirp->d_type == DT_REG)
	{
		returnFile = currentPath + STR("/") + d->dirp->d_name;
		++amount;
		return true;
	}
#endif
}

// Is in the exclude list
bool Files::IsExcluded()
{
	FD d = (FD)data;
	STRING p = d->GetFile();
	for (vector<STRING>::iterator i = exclude.begin(); i != exclude.end(); ++i)
	{
		// Make smaller as it makes things easier
		STRING n = p.substr(0, i->length());
		if (p == (*i))
		//if (STRCMP(p.c_str(), i->c_str()) == 0)
		{
			return true;
		}
	}
	return false;
}

// Check if a directory is open
bool Files::IsDirOpen()
{
	FD d = (FD)data;
#ifdef WIN32
	return (d->find != INVALID_HANDLE_VALUE);
#elif defined(UNIX)
	return (!!d->dp);
#endif
	return false;
}
