

#include "Parameters.hpp"

// Dependies

#include <string>
#include <sstream>

#include <fstream>

#include <math.h>

#include "PixelMap.hpp"
#include "DataValues.hpp"
//#include "Computer.hpp"
#include "System.hpp"

// Parse a string to a list
std::list<int> str2list(const char * str, const int size)
{
	using namespace std;
	list<int> li;

	for (int i = 0, l = 0; true; ++i)
	{
		// Found space or end of line
		if (i >= size || str[i] == ' ')
		{
			string s;
			s.assign(str+l, str+i);
			li.push_back(atoi(s.c_str()));
			l = i+1;
			if (i >= size)
				break;
		}
	}

	return li;
}

// Parse a list to a string
std::string list2str(std::list<int> li)
{
	using namespace std;
	stringstream str;
	for (list<int>::iterator i = li.begin(); i != li.end(); ++i)
	{
		if (i == li.begin())
			str << (*i);
		else
			str << ' ' << (*i);
	}
	return str.str();
}

Parameters::Parameters()
: pxlmp(0)
{
	Reset();
}

// Clear arguments for next use
void Parameters::Reset()
{
	slice.clear();
	height = 0;
	worldVersion = BETA;
	map = World();
	map.n = 1;
	map.s = L"World1";
	night = false;
	noRender = false;
	caveMode = false;
	dimension = 0;
	noHeightMap = false;
	interpolation = false;
	threads = 1;
#ifndef ANGLED_VIEW
	rotation = 0;
#else
	yaw = 0;
	pitch = 0;
#endif
	nogui = false;
	silent = false;
	autoClose = false;
	openFolder = false;
	debug = false;

	args.Reset();
}

// Command line arguments
void Parameters::DoArgs(wchar_t ** commands, int nArgs)
{
	if (!pxlmp)
		return;
	if (nArgs <= 1)
		return;
	using namespace std;

	// Parameters
	Reset();

	// Create object
	//ParseArguments args = ParseArguments();

	// Predefine
	//UINT arguments[32] = {0};

	// Initialization
	/*arguments[0] = args.AddParameter(L'w', PARAM_WSTRING); // World folder
	arguments[1] = args.AddParameter(L'f', PARAM_WSTRING); // Output file
	arguments[2] = args.AddParameter(L'l', PARAM_WSTRING); // Log file
	{
		wchar_t temp2[] = {
			'w', WATER, 
			'w', STATIONARYWATER, 
			'l', LAVA, 
			'l', STATIONARYLAVA, 
			't', LOG, 
			't', LEAVES, 
			0};
		arguments[3] = args.AddParameter(L'o', PARAM_INT | PARAM_LIST, temp2); // Only mode
		arguments[4] = args.AddParameter(L'x', PARAM_INT | PARAM_LIST, temp2); // Exclude mode
	}
	arguments[5] = args.AddParameter(L'n', PARAM_BOOL); // Night mode
	arguments[6] = args.AddParameter(L's', PARAM_INT | PARAM_LIST); // Slice mode
	arguments[7] = args.AddParameter(L'h', PARAM_STRING); // Height mode
	arguments[8] = args.AddParameter(L't', PARAM_INT); // Threads
	arguments[9] = args.AddParameter(L'c', PARAM_BOOL); // Cave mode
	arguments[14] = args.AddParameter(L'r', PARAM_INT); // Rotation mode
	arguments[10] = args.AddParameter(L"texture", PARAM_WSTRING); // Texture mode
	arguments[11] = args.AddParameter(L"no-render", PARAM_BOOL); // No render mode
	arguments[12] = args.AddParameter(L"nether", PARAM_BOOL); // Nether mode
	arguments[15] = args.AddParameter(L"no-gui", PARAM_BOOL); // No GUI mode
	arguments[16] = args.AddParameter(L"silent", PARAM_BOOL); // Silent mode
	arguments[17] = args.AddParameter(L"cache", PARAM_WSTRING); // Cache file
	arguments[18] = args.AddParameter(L"no-heightmap", PARAM_BOOL); // Disable heightmap
	arguments[19] = args.AddParameter(L"auto-close", PARAM_BOOL); // Closes application after rendering
	arguments[20] = args.AddParameter(L"open-folder", PARAM_BOOL); // Opens folder where saved
	arguments[21] = args.AddParameter(L"version", PARAM_STRING); // Set world version
	arguments[22] = args.AddParameter(L"debug", PARAM_BOOL); // Set debug options*/
	AddParameter(L'w', PARAM_WSTRING); // World folder
	AddParameter(L'f', PARAM_WSTRING); // Output file
	AddParameter(L'l', PARAM_WSTRING); // Log file
	{
		wchar_t temp2[] = {
			'w', WATER, 
			'w', STATIONARYWATER, 
			'l', LAVA, 
			'l', STATIONARYLAVA, 
			't', LOG, 
			't', LEAVES, 
			0};
		AddParameter(L'o', PARAM_INT | PARAM_LIST, temp2); // Only mode
		AddParameter(L'x', PARAM_INT | PARAM_LIST, temp2); // Exclude mode
	}
	AddParameter(L'n', PARAM_BOOL); // Night mode
	AddParameter(L's', PARAM_INT | PARAM_LIST); // Slice mode
	AddParameter(L'h', PARAM_STRING); // Height mode
	AddParameter(L't', PARAM_INT); // Threads
	AddParameter(L'c', PARAM_BOOL); // Cave mode
	AddParameter(L'r', PARAM_INT | PARAM_LIST); // Rotation mode
	AddParameter(L"texture", PARAM_WSTRING); // Texture mode
	AddParameter(L"no-render", PARAM_BOOL); // No render mode
	AddParameter(L"nether", PARAM_BOOL); // Nether mode(Deprecated)
	AddParameter(L"dimension", PARAM_INT); // Dimension
	AddParameter(L"no-gui", PARAM_BOOL); // No GUI mode
	AddParameter(L"silent", PARAM_BOOL); // Silent mode
	AddParameter(L"cache", PARAM_WSTRING); // Cache file
	AddParameter(L"no-heightmap", PARAM_BOOL); // Disable heightmap
	AddParameter(L"auto-close", PARAM_BOOL); // Closes application after rendering
	AddParameter(L"open-folder", PARAM_BOOL); // Opens folder where saved
	AddParameter(L"version", PARAM_STRING); // Set world version
	AddParameter(L"debug", PARAM_BOOL); // Set debug options

	// Parse arguments
	args.PushArguments(commands, nArgs);

	// Version
	GetParamVersion();

	// Get world
	GetParamWorld();

	GetParameter(L'f')->Get(&saveFile);

	GetParamLog();

	GetParameter(L'o')->Get(&only);
	GetParameter(L'x')->Get(&exclude);
	GetParameter(L'n')->Get(&night);
	
	GetParamSlice();

	// Height parameter
	GetParamHeight();

	GetParamThreads();

	GetParameter(L'c')->Get(&caveMode);

	GetParamRotation();

	GetParamDimension();

	GetParameter(L"texture")->Get(&texture);
	GetParameter(L"no-render")->Get(&noRender);
	GetParameter(L"no-gui")->Get(&nogui);
	GetParameter(L"silent")->Get(&silent);
	
	GetParamCache();

	GetParameter(L"no-heightmap")->Get(&noHeightMap);
	GetParameter(L"auto-close")->Get(&autoClose);
	GetParameter(L"open-folder")->Get(&openFolder);
	GetParameter(L"debug")->Get(&debug);

	if (nogui)
		autoClose = true;

	// Only one of them
	if (!only.empty())
		exclude.clear();

	// Set debug file
	if (debug)
		pxlmp->debugFile.Open(Port::ReplacePath(STR("./debug.txt")));
}

#ifdef WIN32
// Go through all gui controls
void Parameters::DoArgs(GUI_Creator * gui)
{
	if (!pxlmp)
		return;
	if (!gui)
		return;
	using namespace std;

	// Parameters
	Reset();

	BOOL b;

	GetGUIVersion(gui);

	GetGUIWorld(gui);

	GetGUIDimension(gui);

	// Savefile
	gui->Get(GUI_OUTPUT, saveFile);

	// Texture
	gui->Get(GUI_TEXTURE, texture);

	// Cache
	GetGUICache(gui);

	// Log
	GetGUILog(gui);

	// Bools
	night = !!gui->Get(GUI_NIGHT, b);
	noRender = !!gui->Get(GUI_NORENDER, b);
	caveMode = !!gui->Get(GUI_CAVEMODE, b);
	autoClose = !!gui->Get(GUI_AUTO_CLOSE, b);
	openFolder = !!gui->Get(GUI_OPEN_FOLDER, b);
	noHeightMap = !!gui->Get(GUI_NOHEIGHTMAP, b);
	interpolation = !!gui->Get(GUI_INTERPOLATION, b);

	// Rotation
	GetGUIRotation(gui);

	GetGUIHeight(gui);

	// Threads
	GetGUIThreads(gui);

	// Slice
	GetGUISlice(gui);

	// Only and exclude, only one of them
	GetGUIOnlyExclude(gui);

	gui->Get(GUI_TEXTURE, texture);
}

// Set the args for the GUI
void Parameters::GetArgs(GUI_Creator * gui)
{
	if (!pxlmp)
		return;
	if (!gui)
		return;
	using namespace std;

	gui->Set(GUI_AUTO, worldVersion == DISCOVER);
	gui->Set(GUI_ALPHA, worldVersion == ALPHA);
	gui->Set(GUI_BETA, worldVersion == BETA);
	gui->Set(GUI_ANVIL, worldVersion == ANVIL);

	bool worldfile = map.s.find_first_of(L'/') == -1 && map.s.find_first_of(L'\\') == -1;
	gui->Set(GUI_WORLD, worldfile);
	gui->Set(GUI_CUSTOM, !worldfile);
	gui->Enable(GUI_WORLDFILE, worldfile);
	gui->Enable(GUI_CUSTOMFILE, !worldfile);
	if (worldfile)
		gui->Set(GUI_WORLDFILE, map.s);
	else
		gui->Set(GUI_CUSTOMFILE, map.s);

	// Savefile
	gui->Set(GUI_OUTPUT, saveFile);

	// Texture
	gui->Set(GUI_TEXTURE, texture);

	// Cache
	if (!cache.empty())
	{
		gui->Set(GUI_CACHE, TRUE);
		gui->Enable(GUI_CACHEFILE, TRUE);
		gui->Set(GUI_CACHEFILE, cache);
	}

	// Log
	if (!logFile.empty())
	{
		gui->Set(GUI_LOG, TRUE);
		gui->Enable(GUI_LOGFILE, TRUE);
		gui->Set(GUI_LOGFILE, logFile);
	}

	gui->Set(GUI_DIMENSION, STRN(dimension));

	// Bools
	gui->Set(GUI_NIGHT, night);
	gui->Set(GUI_NORENDER, noRender);
	gui->Set(GUI_CAVEMODE, caveMode);
	gui->Set(GUI_AUTO_CLOSE, autoClose);
	gui->Set(GUI_OPEN_FOLDER, openFolder);
	gui->Set(GUI_NOHEIGHTMAP, noHeightMap);
	gui->Set(GUI_INTERPOLATION, interpolation);

	// Rotation
#ifndef ANGLED_VIEW
	gui->Set(GUI_ROTATE, rotation / 90);
#endif

	gui->Set(GUI_NORMAL, FALSE);
	gui->Set(GUI_FREQ, FALSE);
	gui->Set(GUI_GRAY, FALSE);
	gui->Set(GUI_COLOR, FALSE);
	gui->Set(GUI_SOLID, FALSE);
	gui->Enable(GUI_FREQ_H, FALSE);
	// Modes
	switch (height)
	{
		case 0:
			gui->Set(GUI_NORMAL, TRUE);
			break;
		case -1:
			gui->Set(GUI_GRAY, TRUE);
			break;
		case -2:
			gui->Set(GUI_COLOR, TRUE);
			break;
		case -3:
			gui->Set(GUI_SOLID, TRUE);
			break;
		default:
			gui->Set(GUI_FREQ, TRUE);
			gui->Enable(GUI_FREQ_H, TRUE);
			gui->Set(GUI_FREQ_H, STRN(height));
	}

	// Threads
	if (threads > 1)
	{
		stringstream str;
		str << threads;
		gui->Set(GUI_THREADS, str.str());
		gui->Set(GUI_USETHREADS, TRUE);
		gui->Enable(GUI_THREADS, TRUE);
	}
	else
	{
		gui->Set(GUI_USETHREADS, FALSE);
		gui->Enable(GUI_THREADS, FALSE);
	}

	// Slice
	{
		if (slice.size() > 0)
		{
			gui->Set(GUI_SLICE, STRN(slice.front()));
		}
		if (slice.size() > 1)
		{
			gui->Set(GUI_SLICE_SINGLE, slice.back());
		}
	}

	// Init block colors
	if (!texture.empty())
		gui->Set(GUI_TEXTURE, texture);

	// Only
	if (only.size())
	{
		gui->Set(GUI_ONLY, list2str(only));
	}
	else if (exclude.size())
	{
		gui->Set(GUI_EXCLUDE, list2str(exclude));
	}
}

#endif // WIN32

// Prepare arguments
void Parameters::PrepareArgs()
{
	using namespace std;

	// Add parameters
	LevelPrefs prefs = LevelPrefs();

	// Init block colors
	PrepareColor();

	// Pick world version
	prefs.version = worldVersion;

	STRINGSTREAM str;

	// Create output file with variables
	STRING params;
	// Slice
	params += PrepareSlice(prefs);

	params += PrepareHeight(prefs);

	// Night
	params += PrepareNight(prefs);
	// Cavemode
	params += PrepareCave(prefs);
	// Dimension
	params += PrepareDimension(prefs);
	// Rotation
	params += PrepareRotation(prefs);
	// No rendering
	if (noRender)
		prefs.flags |= CHUNKP_NORENDER;

	prefs.threads = (BYTE)threads;
	// Add log file
	PrepareLog(prefs);
	// Add cache folder
	PrepareCache(prefs);
	// Use only these blocks
	PrepareOnlyExclude(prefs);
	// Special savefile
	pxlmp->save.assign(PrepareSaveFile(params));
	// Disable heightmap
	if (noHeightMap)
		prefs.flags |= CHUNKP_NOHEIGHTMAP;
	// Interpolation
	if (interpolation)
		prefs.flags |= CHUNKP_INTERPOLATION;

	// Validate folders
	pxlmp->folder = Port::ReplacePath(pxlmp->folder);
	pxlmp->save = Port::ReplacePath(pxlmp->save);

	pxlmp->level->prefs = prefs;
}

// Prepare map path
void Parameters::PrepareMap()
{
	PrepareMapWorld();

	PrepareMapSaveFile();
}

// Methods for parameters

UINT Parameters::AddParameter(char p, int type, const char * l)
{
	UINT num = args.AddParameter(p, type, l);
	arguments.push_back(num);
	return num;
}

UINT Parameters::AddParameter(wchar_t p, int type, const wchar_t * l)
{
	UINT num = args.AddParameter(p, type, l);
	arguments.push_back(num);
	return num;
}

UINT Parameters::AddParameter(const char * p, int type, const char * l)
{
	UINT num = args.AddParameter(p, type, l);
	arguments.push_back(num);
	return num;
}

UINT Parameters::AddParameter(const wchar_t * p, int type, const wchar_t * l)
{
	UINT num = args.AddParameter(p, type, l);
	arguments.push_back(num);
	return num;
}

Parameter * Parameters::GetParameter(char p)
{
	return args.GetParameter(p);
}

Parameter * Parameters::GetParameter(wchar_t p)
{
	return args.GetParameter(p);
}

Parameter * Parameters::GetParameter(const char * p)
{
	return args.GetParameter(p);
}

Parameter * Parameters::GetParameter(const wchar_t * p)
{
	return args.GetParameter(p);
}

// Specialized methods for parameters
using namespace std;

void Parameters::GetParamVersion()
{
	string t;
	if (GetParameter(L"version")->Get(&t) && !t.empty())
	{
		// Make lowercase
		for (UINT i = 0; i < t.size(); ++i)
			t[i] = tolower(t[i]);
		// Alpha
		if (t.compare("alpha") == 0)
		{
			worldVersion = ALPHA;
		}
		else if (t.compare("beta") == 0)
		{
			worldVersion = BETA;
		}
		else
		{
			worldVersion = DISCOVER;
		}
	}
}

void Parameters::GetParamWorld()
{
	wstring temp;
	if (GetParameter(L'w')->Get(&temp))
	{
		if (worldVersion == ALPHA)
		{
			if (temp.size())
				map = temp;
			if (map < 0 || map > 5)
				map = 1;
			if (map > 0 && map <= 5)
			{
				map.s = L"World" + map.s;
			}
		}
		else
		{
			map = temp;
		}
	}
}

void Parameters::GetParamDimension()
{
	bool nether;
	// Backwards compatibility
	if (GetParameter(L"nether")->Get(&nether))
	{
		dimension = -1;
	}
	// Default dimensions
	else if (!GetParameter(L"dimension")->Get(&dimension))
	{
		dimension = 0;
	}
}

void Parameters::GetParamLog()
{
	if (GetParameter(L'l')->Get(&logFile) && logFile.empty())
		logFile = L"log.txt";
}

void Parameters::GetParamSlice()
{
	if (GetParameter(L's')->Get(&slice))
	{
		if (slice.size() > 2)
			slice.resize(2);
		if (slice.size() >= 1 && (slice.front() < 0 || slice.front() > 127))
		{
			slice.clear();
			slice.push_back(-1);
		}
	}
}

void Parameters::GetParamHeight()
{
	wstring temp;
	GetParameter(L'h')->Get(&temp);
	height = _wtoi(temp.c_str());
	if (height == 0)
	{
		wchar_t t = temp[0];
		switch (t)
		{
			// Grey
			case L'g':
				height = -1;
				break;
			// Color
			case L'c':
				height = -2;
				break;
			// Solid
			case L's':
				height = -3;
				break;
		}
	}
	else
	{
		// Limit
		if (height <= 1 || height > MAPZ / 8)
			height = 0;
	}
}

void Parameters::GetParamThreads()
{
	if (GetParameter(L't')->Get((int *)&threads) && threads == 0)
		// Automatic detection
		threads = Port::GetLogicalCores();
	// Failproof
	if (threads < 1 || threads > 255) // 255: You don't need more anyway
		threads = 1;
}

void Parameters::GetParamRotation()
{
#ifndef ANGLED_VIEW
	if (!GetParameter(L'r')->Get(&rotation))
		rotation = 0;
#else
	std::list<int> rot;
	if (GetParameter(L'r')->Get(&rot) && rot.size() > 0)
	{
		if (rot.size() > 2)
			rot.resize(2);
		// Yawning
		if (rot.size() == 1)
		{
			yaw = rot.front() % 360;
			pitch = 0;
		}
		// Pitching
		else if (rot.size() == 2)
		{
			yaw = rot.front() % 360;
			if (rot.back() < 0 || rot.back() > 180)
				pitch = 0;
			else
				pitch = rot.back();
		}
	}
	// No rotation
	else
	{
		yaw = 0;
		pitch = 0;
	}
#endif
}

void Parameters::GetParamCache()
{
	if (GetParameter(L"cache")->Get(&cache) && cache.empty())
	{
		// Default cache
		STRING str = pxlmp->GetTemp();
		cache.assign(str.begin(), str.end());
	}
}

// Specialized methods for GUI

#ifdef WIN32
void Parameters::GetGUIVersion(GUI_Creator * gui)
{
	BOOL b;

	if (gui->Get(GUI_ALPHA, b))
	{
		worldVersion = ALPHA;
	}
	else if (gui->Get(GUI_BETA, b))
	{
		worldVersion = BETA;
	}
	else if (gui->Get(GUI_AUTO, b))
	{
		worldVersion = DISCOVER;
	}
	else if(gui->Get(GUI_ANVIL, b))
	{
		worldVersion = ANVIL;
	}
}

void Parameters::GetGUIWorld(GUI_Creator * gui)
{
	BOOL b;

	// Worlds from minecraft save folder
	if (gui->Get(GUI_WORLD, b))
	{
		wstring temp;
		//gui->Get(GUI_WORLDFILE, temp);
		int i = 0;
		gui->GetData(GUI_WORLDFILE, &i);
		STRING temp2 = pxlmp->worlds[i];
		temp.assign(temp2.begin(), temp2.end());
		map = temp;
		if (worldVersion == ALPHA)
		{
			if (map < 0 || map > 5)
				map = 1;
			if (map > 0 && map <= 5)
			{
				map = L"World" + map.s;
			}
		}
	}
	// Custom world(Probably server)
	else if (gui->Get(GUI_CUSTOM, b))
	{
		wstring temp;
		gui->Get(GUI_CUSTOMFILE, temp);
		map = temp;
	}
}

void Parameters::GetGUIDimension(GUI_Creator * gui)
{
	wstring dim;
	if (gui->Get(GUI_DIMENSION, dim))
	{
		dimension = _wtoi(dim.c_str());
	}
	else
	{
		dimension = 0;
	}
}

void Parameters::GetGUICache(GUI_Creator * gui)
{
	BOOL b;

	if (gui->Get(GUI_CACHE, b) && gui->Get(GUI_CACHEFILE, cache) == 0)
	{
		// Default cache
		STRING str = pxlmp->GetTemp();
		cache.assign(str.begin(), str.end());
	}
}

void Parameters::GetGUILog(GUI_Creator * gui)
{
	BOOL b;

	if (gui->Get(GUI_LOG, b) && gui->Get(GUI_LOGFILE, logFile) == 0)
		logFile = L"log.txt";
}

void Parameters::GetGUIRotation(GUI_Creator *gui)
{
#ifdef ANGLED_VIEW
	int num = 0;
	gui->GetData(GUI_ROTATION, &num);
	// Top-down
	if (num == ROT_TOPDOWN)
	{
		yaw = 0;
		pitch = 0;
	}
	// Custom  rotation
	else if (num == ROT_CUSTOM)
	{
		STRING s;
		int t = 0;
		gui->Get(GUI_YAW, s);
		t = STOI(s);
		t = t % 360;
		yaw = degToRad(t);
		gui->Get(GUI_PITCH, s);
		t = STOI(s);
		t = t % 180;
		pitch = degToRad(t);
	}
	// Specified rotation
	else
	{
		// 45 degrees angled view
		if (num & ROT_ANGLED)
			pitch = sin(degToRad(45));
		// Sort of yaw
		if (num & ROT_OBLIQUE)
			yaw = sin(degToRad(45));
		else if (num & ROT_ISOMETRIC)
			yaw = atan(sin(degToRad(45)));
	}
#else
	wstring temp;
	gui->Get(GUI_ROTATE, temp);
	rotation = _wtoi(temp.c_str());
#endif
}

void Parameters::GetGUIHeight(GUI_Creator * gui)
{
	BOOL b;

	// Modes
	// Normal
	if (gui->Get(GUI_NORMAL, b))
		height = 0;
	// Height
	else if (gui->Get(GUI_FREQ, b))
	{
		wstring temp;
		gui->Get(GUI_FREQ_H, temp);
		height = _wtoi(temp.c_str());
		// Limit
		if (height <= 1 || height > MAPZ / 8)
			height = 0;
	}
	// Gray
	else if (gui->Get(GUI_GRAY, b))
		height = -1;
	// Color
	else if (gui->Get(GUI_COLOR, b))
		height = -2;
	// Solid
	else if (gui->Get(GUI_SOLID, b))
		height = -3;
}

void Parameters::GetGUIThreads(GUI_Creator * gui)
{
	BOOL b;

	if (gui->Get(GUI_USETHREADS, b))
	{
		wstring temp;
		int thr = 0;
		// Manual threads
		if (gui->Get(GUI_THREADS, temp))
			thr = _wtoi(temp.c_str());
		// Automatic threads
		if (thr < 1)
			thr = Port::GetLogicalCores();
		threads = thr;
	}
	else
	{
		threads = 1;
	}
}

void Parameters::GetGUISlice(GUI_Creator * gui)
{
	wstring temp;
	int tt = -1;
	if (gui->Get(GUI_SLICE, temp))
		tt = _wtoi(temp.c_str());
	if (tt < 0 || tt > MAPZ)
		tt = -1;
	slice.push_back(tt);
	gui->Get(GUI_SLICE_SINGLE, tt);
	slice.push_back(tt);
}

void Parameters::GetGUIOnlyExclude(GUI_Creator * gui)
{
	std::string str;
	int size = 0;
	// Only
	if (size = gui->Get(GUI_ONLY, str))
	{
		only = str2list(str.c_str(), size);
	}
	// Exclude
	else if (size = gui->Get(GUI_EXCLUDE, str))
	{
		exclude = str2list(str.c_str(), size);
	}
}
#endif // WIN32

// Specialized methods for preparing map

void Parameters::PrepareMapWorld()
{
	// Define world folder
	STRING home = pxlmp->GetSaveFolder();
	// Classic
	if (worldVersion == CLASSIC)
	{
		return;
	}
	// Invdev/Infdev
	else if (worldVersion == INVDEV)
	{
		return;
	}
	// Alpha
	else if (worldVersion == ALPHA)
	{
		if (map.n != 0)
		{
			pxlmp->world.assign(map.s.begin(), map.s.end());
			pxlmp->folder = home + STR("/") + pxlmp->world;
		}
		else
		{
			pxlmp->folder.assign(map.s.begin(), map.s.end());
		}
	}
	// Beta
	else if (worldVersion == BETA || worldVersion == ANVIL)
	{
		map = Port::Replace(map.s, L'\\', L'/');
		if (map.s.size())
		{
			if (map.s.find_first_of(L'/') == -1)
			{
				pxlmp->world.assign(map.s.begin(), map.s.end());
				pxlmp->folder = home + STR("/") + pxlmp->world;
			}
			else
			{
				pxlmp->folder.assign(map.s.begin(), map.s.end());
			}
			pxlmp->folder.assign(pxlmp->folder.c_str());
		}
		else
		{
			pxlmp->go.error = LVL_ERR_INVALIDWORLD;
			return;
		}
	}
	// Dimension
	if (dimension != 0)
	{
		pxlmp->folder.append(STR("/DIM") + STRN(dimension));
	}
	// McRegion
	if (worldVersion == BETA)
	{
		pxlmp->folder.append(STR("/region"));
	}
}

void Parameters::PrepareMapSaveFile()
{
	// Specified save file
	if (saveFile.size() > 0)
	{
		saveFile = Port::Replace(saveFile, L'/', L'\\');
		pxlmp->save.assign(saveFile.begin(), saveFile.end());
	}
	// Use world folder name
	else if (map.n == 0)
	{
		wstring temp = map.s;
		size_t num = temp.find_last_of(L'/');
		if (num >= 0)
			temp = temp.substr(num+1);
		else
		{
			num = temp.find_last_of(L'/');
			if (num >= 0)
				temp = temp.substr(num+1);
		}
		pxlmp->save.assign(STR("./"));
		pxlmp->save.append(temp.begin(), temp.end());
	}
	// Default worlds
	else
	{
		pxlmp->save.assign(STR("./") + pxlmp->world);
	}
}

// Specalized methods for preparing arguments

void Parameters::PrepareColor()
{
	if (texture.empty())
		DataValues::MakeColor();
	else
	{
		STRING t;
		t.assign(texture.begin(), texture.end());
		DataValues::MakeColor(t);
	}
}

STRING Parameters::PrepareSlice(LevelPrefs & prefs)
{
	STRING str;
	if (slice.size() && slice.front() != -1)
	{
		prefs.z = (BYTE)slice.front();
		//str << STR("-slice") << (int)prefs.z;
		str += STR("-slice") + STRN((int)prefs.z);
		if (slice.size() == 2 && slice.back() != 0)
		{
			prefs.flags |= CHUNKP_SINGLESLICE;
			str += STR("S");
			//str << STR("S");
		}
	}
	return str;
}

STRING Parameters::PrepareHeight(LevelPrefs & prefs)
{
	STRING str;
	// Line heightmap
	if (height > 0)
	{
		prefs.h = (BYTE)height;
		//str << STR("-height") << height;
		str += STR("-height") + STRN(height);
	}
	// Gray heightmap
	else if (height == -1)
	{
		prefs.flags |= CHUNKP_GRAY;
		//str << STR("-gray");
		str += STR("-gray");
	}
	// Color heightmap
	else if (height == -2)
	{
		prefs.flags |= CHUNKP_COLOR;
		//str << STR("-color");
		str += STR("-color");
	}
	// Solid heightmap
	else if (height == -3)
	{
		prefs.flags |= CHUNKP_SOLID;
		//str << STR("-solid");
		str += STR("-solid");
	}
	return str;
}

STRING Parameters::PrepareNight(LevelPrefs & prefs)
{
	STRING str;
	if (night)
	{
		prefs.flags |= CHUNKP_NIGHT;
		str += STR("-night");
	}
	return str;
}

STRING Parameters::PrepareCave(LevelPrefs & prefs)
{
	STRING str;
	if (caveMode)
	{
		prefs.flags |= CHUNKP_CAVE;
		str += STR("-cave");
	}
	return str;
}

STRING Parameters::PrepareDimension(LevelPrefs & prefs)
{
	STRING str;
	if (dimension)
	{
		// Cave is only because nether is below surface
		prefs.flags |= CHUNKP_CAVE;
		str += STR("-dim-") + STRN(dimension);
		//str << STR("-nether");
		//pxlmp->folder.append(STR("\\DIM-1"));
	}
	return str;
}

STRING Parameters::PrepareRotation(LevelPrefs & prefs)
{
	STRING str;
	prefs.rot.width = MAPX;
	prefs.rot.height = MAPY;
	prefs.rot.depth = MAPZ;
#ifdef ANGLED_VIEW
	{
		if (!(yaw == 0 && pitch == 0))
			//str << STR("-angled");
			str += STR("-angled");
		prefs.rot.Transform(yaw, pitch);
		prefs.yaw = yaw;
		prefs.pitch = pitch;
	}
#else
	if (rotation != 0)
	{
		prefs.rotation = rotation;
		//str << STR("-rot") << rotation;
		str += STR("-rot") + STRN(ROTATION);
	}
#endif
	return str;
}

void Parameters::PrepareLog(LevelPrefs & prefs)
{
	prefs.log.assign(logFile.begin(), logFile.end());
	prefs.log = Port::ReplacePath(prefs.log);
	int tellLog = Port::TellPath(prefs.log);
	// Path only
	if (tellLog & TELLPATH_DIR)
	{
		// Without dir
		if ((tellLog & TELLPATH_END_SLASH) == 0)
			prefs.log.append(STR("/"));
		prefs.log.append(STR("log.txt"));
		prefs.log = Port::ReplacePath(prefs.log);
	}
}

void Parameters::PrepareCache(LevelPrefs & prefs)
{
	prefs.cache.assign(cache.begin(), cache.end());
	prefs.cache = Port::ReplacePath(prefs.cache);
	int tellCache = Port::TellPath(prefs.cache);
	// Path only
	if (tellCache & TELLPATH_DIR)
	{
		// Without dir
		if ((tellCache & TELLPATH_END_SLASH) == 0)
			prefs.cache.append(STR("/"));
		prefs.cache.append(STR("cache.bin"));
		prefs.cache = Port::ReplacePath(prefs.cache);
	}
}

void Parameters::PrepareOnlyExclude(LevelPrefs & prefs)
{
	if (!only.empty())
	{
		memset(prefs.block, false, DATAVALUES_AMOUNT);
		for (list<int>::iterator i = only.begin(); i != only.end(); ++i)
			if (*i >= 0 && *i < 256)
				prefs.block[*i] = true;
	}
	// Exclude these blocks
	else if (!exclude.empty())
	{
		for (list<int>::iterator i = exclude.begin(); i != exclude.end(); ++i)
			if (*i >= 0 && *i < 256)
				prefs.block[*i] = false;
	}
}

STRING Parameters::PrepareSaveFile(STRING & params)
{
	STRING str;
	// Empty path
	if (saveFile.size() == 0)
	{
		str.assign(pxlmp->save.c_str());
		str.append(params + STR(".png"));
		return str;
	}
	str.assign(saveFile.begin(), saveFile.end());
	int tell = Port::TellPath(str);

	// Complete file
	if (tell & TELLPATH_FILE)
	{
		return str;
	}
	// Path only
	else if (tell & TELLPATH_DIR)
	{
		// Without dir
		if ((tell & TELLPATH_END_SLASH) == 0)
			str.append(STR("/"));
		str.append(STR("world") + params + STR(".png"));
	}

	return str;
}
