

#pragma once
#ifndef PARAMETERS_HPP
#define PARAMETERS_HPP

//#include "Globals.hpp"
#include "String.hpp"
#include "Number.hpp"
#include "Level.hpp"

#include "ParseArguments.hpp"

#ifdef WIN32
#include "GUI.hpp"
#endif

class PixelMap;

class Parameters
{
public:
	Parameters();

	void Reset();

	void DoArgs(wchar_t ** commands, int nArgs);
#ifdef WIN32
	void DoArgs(GUI_Creator * gui);
	void GetArgs(GUI_Creator * gui);
#endif // WIN32
	void PrepareArgs();
	void PrepareMap();
	
	// Flags
	bool silent; // Disable notification of finished rendering
	bool nogui; // Disable gui/text output(Tip: Use with autoClose)
	bool autoClose; // Close after finished
	bool openFolder; // Open image output folder
	bool debug; // Outputs a debug file

	// Parameters
	MAP_VERSION worldVersion; // Version of file saved
	int height; // Sort of mode to render in different types, e.g. color, gray, heightmap
	std::list<int> slice; // First is where to start slice, and second is a bool for single slice or whole slice
	World map; // Which map to render
	std::wstring saveFile; // Where to save the image
	std::list<int> only; // This is the only thing to render
	std::list<int> exclude; // This should be excluded
	bool night; // Nightmode
	bool noRender; // Don't render(Good for logging)
	bool caveMode; // Cave mode(Showing caves underground)
	int dimension; // Sort of dimension
	bool noHeightMap; // Disable the heightmap to increase flexibility but also rendertime
	bool interpolation; // Interpolate the file(Only good for angled mode)
	std::wstring logFile; // Where to save a log(If there is one)
	size_t threads; // Amount of threads used
	std::wstring texture; // Load an external texture file
#ifndef ANGLED_VIEW
	int rotation; // Rotation(0,90,180,270)(Soon deprecated)
#else
	double yaw; // Yaw rotation(Left, Right)
	double pitch; // Pitch rotation(Up, Down)
#endif
	std::wstring cache; // Cache file

private:
	friend class PixelMap;
	PixelMap * pxlmp;

	ParseArguments args;
	std::vector<UINT> arguments;
	UINT AddParameter(char p, int type = PARAM_BOOL, const char * l = 0);
	UINT AddParameter(wchar_t p, int type = PARAM_BOOL, const wchar_t * l = 0);
	UINT AddParameter(const char * p, int type = PARAM_BOOL, const char * l = 0);
	UINT AddParameter(const wchar_t * p, int type = PARAM_BOOL, const wchar_t * l = 0);

	Parameter * GetParameter(char p);
	Parameter * GetParameter(wchar_t p);
	Parameter * GetParameter(const char * p);
	Parameter * GetParameter(const wchar_t * p);

	// Specialized get param methods
	void GetParamVersion();
	void GetParamWorld();
	void GetParamDimension();
	void GetParamLog();
	void GetParamSlice();
	void GetParamHeight();
	void GetParamThreads();
	void GetParamRotation();
	void GetParamCache();

	// Specalized get gui methods
#ifdef WIN32
	void GetGUIVersion(GUI_Creator * gui);
	void GetGUIWorld(GUI_Creator * gui);
	void GetGUIDimension(GUI_Creator * gui);
	void GetGUICache(GUI_Creator * gui);
	void GetGUILog(GUI_Creator * gui);
	void GetGUIRotation(GUI_Creator * gui);
	void GetGUIHeight(GUI_Creator * gui);
	void GetGUIThreads(GUI_Creator * gui);
	void GetGUISlice(GUI_Creator * gui);
	void GetGUIOnlyExclude(GUI_Creator * gui);
#endif // WIN32

	// Specalized prepare map methods
	void PrepareMapWorld();
	void PrepareMapSaveFile();

	// Specialized prepare arguments
	void PrepareColor();
	STRING PrepareSlice(LevelPrefs &);
	STRING PrepareHeight(LevelPrefs &);
	STRING PrepareNight(LevelPrefs &);
	STRING PrepareCave(LevelPrefs &);
	STRING PrepareDimension(LevelPrefs &);
	STRING PrepareRotation(LevelPrefs &);
	void PrepareLog(LevelPrefs &);
	void PrepareCache(LevelPrefs &);
	void PrepareOnlyExclude(LevelPrefs &);
	STRING PrepareSaveFile(STRING & str);
};

#endif // PARAMETERS_HPP