// ========
// Name :: ParseArguments
// Usage :: An argument parser to make it easier to
// parse arguments
// ========

#pragma once
#ifndef PARSEARGUMENTS_HPP
#define PARSEARGUMENTS_HPP

//#include "Globals.hpp"
#include "String.hpp"

//#include <string.h>
#include <vector>
#include <map>
#include <list>

#define PARAM_DEFAULT -1

// Param types
#define PARAM_BOOL 0
#define PARAM_INT 1
#define PARAM_STRING 2
#define PARAM_WSTRING 4
#define PARAM_LIST 8

union SCHAR
{
	char c;
	const char * s;
	wchar_t wc;
	const wchar_t * ws;
};

struct CharBox
{
	CharBox(const CharBox & c) { name = c.name; type = c.type; }
	CharBox(char c) { name.c = c; type = 1; }
	CharBox(const char * s) { name.s = s; type = 2; }
	CharBox(wchar_t wc) { name.wc = wc; type = 3; }
	CharBox(const wchar_t * ws) { name.ws = ws; type = 4; }

	CharBox & operator=(const char & c) { name.c = c; return *this; }
	CharBox & operator=(const char * s) { name.s = s; return *this; }
	CharBox & operator=(const wchar_t & wc) { name.wc = wc; return *this; }
	CharBox & operator=(const wchar_t * ws) { name.ws = ws; return *this; }

	bool operator<(const CharBox & c) const
	{
		switch (type)
		{
			case 1:
				return name.c < c.name.c;
			case 2:
				return ASTRING(name.s) < ASTRING(c.name.s);
			case 3:
				return name.wc < c.name.wc;
			case 4:
				return WSTRING(name.ws) < WSTRING(c.name.ws);
		}
		return false;
	}
	bool operator>(const CharBox & c) const
	{
		switch (type)
		{
			case 1:
				return name.c > c.name.c;
			case 2:
				return ASTRING(name.s) > ASTRING(c.name.s);
			case 3:
				return name.wc > c.name.wc;
			case 4:
				return WSTRING(name.ws) > WSTRING(c.name.ws);
		}
		return false;
	}

	SCHAR name;
	char type;
};

class ParseArguments;

// Parameter
struct Parameter
{
	Parameter();
	Parameter(const Parameter & a);
	// Assign
	Parameter & operator=(const Parameter & a);
	Parameter & operator=(const bool & a);
	Parameter & operator=(const int & a);
	Parameter & operator=(const std::string & a);
	Parameter & operator=(const std::wstring & a);
	Parameter & operator=(const std::list<int> & a);
	Parameter & operator=(const std::list<std::string> & a);
	Parameter & operator=(const std::list<std::wstring> a);
	// Get
	bool Get(bool * a);
	bool Get(int * a);
	bool Get(std::string * a);
	bool Get(std::wstring * a);
	bool Get(std::list<int> * a);
	bool Get(std::list<std::string> * a);
	bool Get(std::list<std::wstring> * a);

	// Exceptions
	std::string eStr;
	std::wstring eWstr;

private:
	friend class ParseArguments;

	// Data
	bool boolean;
	int integer;
	std::string str;
	std::wstring wstr;
	std::list<int> nlist;
	std::list<std::string> slist;
	std::list<std::wstring> wlist;

	// Type
	int type;
	// Name
	SCHAR name;

	bool exist;
};

class ParseArguments
{
public:
	~ParseArguments();

	void Reset();

	unsigned int AddParameter(char p, int type = PARAM_BOOL, const char * l = 0);
	unsigned int AddParameter(wchar_t p, int type = PARAM_BOOL, const wchar_t * l = 0);
	unsigned int AddParameter(const char * p, int type = PARAM_BOOL, const char * l = 0);
	unsigned int AddParameter(const wchar_t * p, int type = PARAM_BOOL, const wchar_t * l = 0);

	Parameter * GetParameter(unsigned int i);
	Parameter * GetAdvancedParameter(unsigned int i);
	Parameter * GetParameter(char i);
	Parameter * GetParameter(wchar_t i);
	Parameter * GetParameter(const char * i);
	Parameter * GetParameter(const wchar_t * i);

	void PushArguments(char ** args, const int argc);
	void PushArguments(wchar_t ** args, const int argc);
private:
	//std::vector<Parameter *> params; // Normal
	//std::vector<Parameter *> aparams; // Advanced
	std::map<CharBox, Parameter *> params;
	std::map<CharBox, Parameter *> aparams;
};

#endif // PARSEARGUMENTS_HPP