
#pragma once
#ifndef PIXELMAP_HPP
#define PIXELMAP_HPP

#include <string>
#include <vector>
//#include "Utility.hpp"
#include "Parameters.hpp"

//#include "Globals.hpp"
#include "System.hpp"
#include "Number.hpp"
#include "Level.hpp"

#ifdef WIN32
#include "GUI.hpp"
#endif

#include "Counter.h"

#include "Debug.hpp"

#ifdef UNIX
#include "parallel.h"
typedef THREAD HANDLE;
#endif

struct GoThread
{
	GoThread()
	{
		memset(this, 0, sizeof(GoThread));
	}
	Level * level;
	STRING * folder;
	STRING * save;
	int error;
	bool running;
};

class PixelMap
{
public:
	PixelMap();
	~PixelMap();
	void Initialize();
	void Shutdown();
	void Start();
	void Stop();
	void Abort();
	std::string GetErrorMessage();
	void DoArgs(wchar_t ** commands, int nArgs);
#ifdef WIN32
	void DoArgs(GUI_Creator * gui);
	void GetArgs(GUI_Creator * gui);
	void GetWorlds(GUI_Creator * gui);
#endif
	bool Update();
	bool Display();

	int GetState();
	int GetPercent();
	size_t GetQueue();
	size_t GetFinished();
	size_t GetChunks();
	int GetTime();
	int GetTimeLeft();
	bool NoGui();
	bool OpenFolder();
	bool AutoClose();
	void Log(STRING text);
	void Log(const char * format, ...);
	std::string GetSave();

	// Variables
	static STRING GetTemp();
	static STRING GetSaveFolder();

private:

	friend class Parameters;
	float elapsed;
	float running;

	Counter counter;

	Level * level;
public:
	static void SaveWorldLog(const Level * level, const STRING & folder);
private:

	GoThread go;
	HANDLE thread;
	static THREAD_FUNCTION(WorkingLevel, data);

	// Debug
	Debug debugFile;

	// Time
	//float processTime;
	float timeNotches[Level::AMOUNT];
	float nextDisplay;
	float stageDisplay;

	/*// Flags
	bool silent;
	bool nogui;
	bool autoClose;
	bool openFolder;
	bool debug;*/

	// Parameters
	Parameters param;
	/*struct Parameters
	{
		MAP_VERSION worldVersion;
		int height;
		std::list<int> slice;
		World map;
		std::wstring saveFile;
		std::list<int> only;
		std::list<int> exclude;
		bool night;
		bool noRender;
		bool caveMode;
		bool nether;
		bool noHeightMap;
		bool interpolation;
		std::wstring logFile;
		size_t threads;
		std::wstring texture;
		int rotation;
		double yaw;
		double pitch;
		std::wstring cache;
	} param;*/

	/*MAP_VERSION worldVersion;
	int height;
	std::list<int> slice;
	World map;
	std::wstring saveFile;
	std::list<int> only;
	std::list<int> exclude;
	bool night;
	bool noRender;
	bool caveMode;
	bool nether;
	bool noHeightMap;
	bool interpolation;
	std::wstring logFile;
	size_t threads;
	std::wstring texture;
	int rotation;
	double yaw;
	double pitch;
	std::wstring cache;*/
	// Temporary variables
	STRING world;
	STRING folder;
	STRING save;
	std::vector<STRING> worlds;

	void ResetArgs();
	void PrepareArgs();
	void PrepareMap();
};

#endif // PIXELMAP_HPP