

#pragma once
#ifndef STRING_HPP
#define STRING_HPP

#include <string.h>
#include <string>
#include <sstream>
#include <stdarg.h>

typedef std::string ASTRING;
typedef std::wstring WSTRING;

#ifdef UNIX
#include <stdlib.h>
// Windows only -> Linux compatible function
inline int _wtoi(const wchar_t * wch)
{
	std::wstring wstr;
	wstr.assign(wch);
	std::string str;
	str.assign(wstr.begin(), wstr.end());
	return atoi(str.c_str());
}
#endif // UNIX

// Unicode
#ifdef UNICODE

typedef std::wstring STRING;
typedef std::wstringstream STRINGSTREAM;
#define STOI(str) _wtoi(str.c_str())

// ANSI
#else // UNICODE

typedef std::string STRING;
typedef std::stringstream STRINGSTREAM;
#define STOI(str) atoi(str.c_str())

#endif // UNICODE

namespace Port
{
	// Make string out of character array
	STRING MakeString(const char * text);
	// Make string out of number
	template<typename NATIVE>
	STRING MakeStringN(NATIVE n)
	{
		STRINGSTREAM str;
		str << n;
		return str.str();
	}
	// Replace something in a string
	WSTRING Replace(WSTRING str, wchar_t from, wchar_t to);
	ASTRING Replace(ASTRING str, char from, char to);
	// Convert from one encoding to another
	ASTRING ConvertEncoding(WSTRING str);
	WSTRING ConvertEncoding(ASTRING str);
	// Get length of character string
	size_t StrLen(const char * str);
	size_t StrLen(const wchar_t * str);
	
	int SPrintf(ASTRING &, const char *, ...);
	int SPrintf(ASTRING &, const char *, va_list);
	int SPrintf(WSTRING &, const wchar_t *, ...);
	int SPrintf(WSTRING &, const wchar_t *, va_list);
}

// Easier to use when handling strings
#define STR(str) Port::MakeString(str)
#define STRN(str) Port::MakeStringN(str)
#define STRC(str) STR(str).c_str()
#define STR_CLEAN(str) str.assign(str.c_str())
#define W2A(str) Port::ConvertEncoding(str)
#define A2W(str) Port::ConvertEncoding(str)

#endif // STRING_HPP