

#pragma once
#ifndef THREADPOOL_HPP
#define THREADPOOL_HPP

#ifdef WIN32
#include <windows.h>
#elif defined(UNIX)
#include "parallel.h"
typedef THREAD HANDLE;
#endif

#include <queue>
#include <vector>

typedef void * WORK;
typedef void * RESULT;
typedef RESULT (*THREAD_FUNC)(WORK order);

struct Thread;

class ThreadPool
{
public:
	ThreadPool();
	~ThreadPool();

	void Initialize(int amount, THREAD_FUNC func);
	void Shutdown();

	bool Work();
	inline bool StayAlive() {return process;}
	// Deprecated
	inline bool Synchronized()
	{
		// Not trustworthy
		return status.works == status.results;
	}

	// Work and result
private:
	bool PushWork(WORK order);
	bool FullWork();
	bool EmptyWork();
	bool EmptyResult();
	bool PopResult(RESULT * result);
public:
	// Data handling in temporary buffer
	void Push(WORK order);
	bool Pop(RESULT * result);

	inline void MarkThreadClosed() {++closed;}
	inline int GetClosedThreads() {return closed;}
	inline bool IsRunning() {return (group.size() - closed) > 0;}
private:
	Thread * threads;
	bool volatile process;
	std::queue<WORK> work;
	std::queue<RESULT> result;
	int volatile closed, // Closed threads
		stacked; // Stacked work in pool
	// Statistics
	struct
	{
		int works,
			results;
	} status;

	// Replace
	std::vector<HANDLE> group;
};

#endif // THREADPOOL_HPP