
// Transform functionality for images and chunks

#pragma once
#ifndef TRANSFORM_HPP
#define TRANSFORM_HPP

const double PI = 3.14159265;

#define degToRad(deg) ((deg)*(PI/180))

// Note: Vector3D and Matrix4x4 is copied from a old project, 
// so some methods is not necessary

// 3D vector
struct Vector3D
{
	Vector3D() : x(0), y(0), z(0) {};
	Vector3D(double x, double y, double z) : x(x), y(y), z(z) {}
	double x, y, z;

	// Numerous of overloaded operators
	const Vector3D operator+(const Vector3D u) const;
	const Vector3D operator-(const Vector3D u) const;
	Vector3D operator-() const;
	const Vector3D operator*(const double k) const;
	const Vector3D operator/(const double k) const;

	// Make a dot product
	static double DotProduct(Vector3D v, Vector3D u);

	// Make a cross product
	static Vector3D CrossProduct(Vector3D v, Vector3D u);

	// Get the angle from two vectors
	static double Angle(Vector3D v, Vector3D u);

	// Get closest position
	static bool Compare(Vector3D v, Vector3D u);

	static bool Compare(Vector3D v, Vector3D u, double limit);

	// Crop the float to between -1 and 1
	static double Clamp(double f);

	// Get the length of the vector
	double Length();

	// Normalizes a vector
	Vector3D Normalize();
};

// Matrixes
class Matrix4x4
{
public:
	Matrix4x4(double null = 1);

	Matrix4x4 translate(double x, double y, double z);
	Matrix4x4 scale(double x, double y, double z);
	Matrix4x4 rotX(double degrees);
	Matrix4x4 rotY(double degrees);
	Matrix4x4 rotZ(double degrees);

	Matrix4x4 operator*(Matrix4x4 b);
	double& operator()(int i, int j) {return matrix[i][j];}
	double operator()(int i, int j) const {return matrix[i][j];}
	Vector3D operator*(Vector3D a);
	Matrix4x4 operator+(Matrix4x4 b);

private:
	double matrix[4][4];
};

typedef Matrix4x4 MATRIX;

// Rotation information
struct Rotation
{
	Rotation() : yCos(1), ySin(0), pCos(1), pSin(0), width(0), height(0), depth(0) {}

	void Transform(double yaw, double pitch);
	Vector3D Place(double x, double y, double z);
	bool IsStraight();

	double yCos;
	double ySin;
	double pCos;
	double pSin;
	unsigned int width;
	unsigned int height;
	unsigned int depth;
	MATRIX transform;
};

#endif // TRANSFORM_HPP